/** 

SFSEXP: Small, Fast S-Expression Library version 1.0
Written by Matthew Sottile (matt@lanl.gov)

Copyright (2004). The Regents of the University of California. This material
was produced under U.S. Government contract W-7405-ENG-36 for Los Alamos
National Laboratory, which is operated by the University of California for
the U.S. Department of Energy. The U.S. Government has rights to use,
reproduce, and distribute this software. NEITHER THE GOVERNMENT NOR THE
UNIVERSITY MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LIABILITY
FOR THE USE OF THIS SOFTWARE. If software is modified to produce derivative
works, such modified software should be clearly marked, so as not to confuse
it with the version available from LANL.

Additionally, this program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version. Accordingly, this program is distributed in the
hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

LA-CC-04-094

**/

#include <sys/fcntl.h>
#include "sexp.h"
#include <assert.h>
#include <unistd.h>
#include <stdio.h>

/**
 * read a binary file from fd.  Length is filled in, bss is buffer
 * start size and bgs is buffer growth size for realloc()ing.  returns
 * pointer to byte-buffer.  if NULL, then there was some sort of error.
 */
char *readfile(int fd, int *length, int bss, int bgs) {
  char *buf, *tmp;
  int bused, ballocd;
  int amt;

  /* check for either errors on the caller side, or just plain old stupid
   * people. 
   */
  assert(fd > 0);
  assert(bss > 0);
  assert(bgs > 0);

  bused = 0;
  ballocd = bss;

  buf = (char *)malloc(sizeof(char)*bss);
  assert(buf != NULL);

  while ((amt = read(fd,buf+bused,ballocd-bused)) > 0) {
    bused += amt;
    if (bused == ballocd) {
      tmp = (char *)realloc(buf,ballocd+bgs);
      assert(tmp != NULL);
      buf = tmp;
      ballocd += bgs;
    }
  }

  if (bused == 0) return NULL;

  *length = bused;
  return buf;
}

int main(int argc, char **argv) {
  sexp_t *sx_in, *sx_out;
  int fd;
  char *b;
  int l;
  CSTRING *s;
  pcont_t *pc;

  /* read data */
  fd = open("testdata",O_RDONLY);
  if (fd <= 0) {
    printf("Error opening test data file ``testdata''\n");
    exit(1);
  }
  b = readfile(fd,&l,1024,256);
  close(fd);
  
  /* report */
  printf("Read %d bytes of data.\n",l);

  sx_in = (sexp_t *)malloc(sizeof(sexp_t));
  assert(sx_in != NULL);
  sx_in->ty = SEXP_VALUE;
  sx_in->next = sx_in->list = NULL;
  sx_in->aty = SEXP_BINARY;
  sx_in->bindata = b;
  sx_in->binlength = l;
  sx_in = new_sexp_list(sx_in);

  printf("Created expression.\n");

  print_sexp_cstr(&s,sx_in,l+1024,1024);
  destroy_sexp(sx_in);

  free(b);
  b = NULL;
  sx_in = NULL;

  printf("Destroyed AST and buffer.\n");

  pc = init_continuation(NULL);
  pc->mode = PARSER_INLINE_BINARY;
  pc = cparse_sexp(s->base,s->len,pc);
  sx_out = pc->last_sexp;

  printf("Parsed unparsed version back to AST.\n");

  assert(sx_out != NULL);

  b = sx_out->list->bindata;
  l = sx_out->list->binlength;

  fd = open("testdata_out",O_RDWR);
  if (fd <= 0) {
    printf("Error opening ``testdata_out'': Create empty file to write to.\n");
    exit(1);
  }
  write(fd,b,l);
  close(fd);

  printf("Extracted and wrote bindata from AST.\n");

  exit(1);
}
